/*

  Bump Earth!

  all textures courtesy of http://www.shadedrelief.com/natural3/index.html

  adapted by Christopher Rasmussen from Texturing tutorial here:
  http://www.lighthouse3d.com/tutorials/glsl-tutorial/shader-examples 

  May, 2012

  Notes:

  main.cpp is pretty much the same as regular Earth main, except the light is moved a bit, upped
  number of vertices in sphere model, elevation texture is default height map, and 'w' (UP) and 's' (DOWN) 
  control bump exaggeration factor.
  
  for some fun, choose water texture ('2') to be height map

*/

//----------------------------------------------------------------------------
//----------------------------------------------------------------------------

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <GL/glew.h>
#include <GL/glut.h>

#include "textfile.h"

//----------------------------------------------------------------------------
//----------------------------------------------------------------------------

float exaggeration_factor = 1.0;

GLuint v,f,p;
float lpos[4] = {5.0, 5.0, 1.0, 0};

GLUquadricObj *qObj;

#define TEXTURE_HEIGHT     1
#define TEXTURE_WATER      2
#define TEXTURE_NIGHT      3
#define TEXTURE_CLOUDS     4
#define TEXTURE_COLOR      5

unsigned char *read_PPM(char *filename, int *width, int *height);

//----------------------------------------------------------------------------
//----------------------------------------------------------------------------

void changeSize(int w, int h) 
{
  
  // Prevent a divide by zero, when window is too short (you can't make a window of zero width).
  if (h == 0)
    h = 1;
  
  float ratio = 1.0 * w / h;
  
  // Reset the coordinate system before modifying
  glMatrixMode(GL_PROJECTION);
  glLoadIdentity();
  
  // Set the viewport to be the entire window
  glViewport(0, 0, w, h);
  
  // Set the correct perspective.
  gluPerspective(45,ratio,1,1000);
  glMatrixMode(GL_MODELVIEW);
  
}

//----------------------------------------------------------------------------

float a = 0;

void renderScene(void) 
{
  glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
  
  glLoadIdentity();
  gluLookAt(0.0, 0.0, 6.0, 
 	    0.0, 0.0, 0.0,
 	    0.0, 1.0, 0.0);
  
  glLightfv(GL_LIGHT0, GL_POSITION, lpos);

  glRotatef(a, 0, 1, 0);   // rotation around axis
  glRotatef(90, 1, 0, 0);  // seemingly necessary to get texture in correct orientation

  gluSphere(qObj, 2.2, 64, 64);

  a += 0.005;

  glFlush();
  glutSwapBuffers();
}

//----------------------------------------------------------------------------

void keyboard(unsigned char key, int x, int y) 
{
  if (key == 'q') 
    exit(0);
  else if (key == 'w') {
    exaggeration_factor += 1.0;
    printf("Exaggeration = %.1f\n", exaggeration_factor);
    glUniform1f(glGetUniformLocation(p, "Exaggeration"), exaggeration_factor);
  }
  else if (key == 's') {
    if (exaggeration_factor >= 2.0)
    exaggeration_factor -= 1.0;
    printf("Exaggeration = %.1f\n", exaggeration_factor);
    glUniform1f(glGetUniformLocation(p, "Exaggeration"), exaggeration_factor);
  }
  else if (key == '1') {
    glBindTexture(GL_TEXTURE_2D, 1);  
    glutPostRedisplay();
  }
  else if (key == '2') {
    glBindTexture(GL_TEXTURE_2D, 2);  
    glutPostRedisplay();
  }
  else if (key == '3') {
    glBindTexture(GL_TEXTURE_2D, 3);  
    glutPostRedisplay();
  }
  else if (key == '4') {
    glBindTexture(GL_TEXTURE_2D, 4);  
    glutPostRedisplay();
  }
  else if (key == '5') {
    glBindTexture(GL_TEXTURE_2D, 5);  
    glutPostRedisplay();
  }
}

//----------------------------------------------------------------------------

#define printOpenGLError() printOglError(__FILE__, __LINE__)

// Returns 1 if an OpenGL error occurred, 0 otherwise.

int printOglError(char *file, int line)
{
  GLenum glErr;
  int    retCode = 0;
  
  glErr = glGetError();
  while (glErr != GL_NO_ERROR)
    {
      printf("glError in file %s @ line %d: %s\n", file, line, gluErrorString(glErr));
      retCode = 1;
      glErr = glGetError();
    }
  return retCode;
}

//----------------------------------------------------------------------------

void printShaderInfoLog(GLuint obj)
{
  int infologLength = 0;
  int charsWritten  = 0;
  char *infoLog;
  
  glGetShaderiv(obj, GL_INFO_LOG_LENGTH,&infologLength);
  
  if (infologLength > 0) {
    infoLog = (char *)malloc(infologLength);
    glGetShaderInfoLog(obj, infologLength, &charsWritten, infoLog);
    printf("%s\n",infoLog);
    free(infoLog);
  }
}

//----------------------------------------------------------------------------

void printProgramInfoLog(GLuint obj)
{
  int infologLength = 0;
  int charsWritten  = 0;
  char *infoLog;
  
  glGetProgramiv(obj, GL_INFO_LOG_LENGTH,&infologLength);
  
  if (infologLength > 0) {
    infoLog = (char *)malloc(infologLength);
    glGetProgramInfoLog(obj, infologLength, &charsWritten, infoLog);
    printf("%s\n",infoLog);
    free(infoLog);
  }
}

//----------------------------------------------------------------------------

void setShaders() 
{
  char *vs = NULL,*fs = NULL,*fs2 = NULL;
  
  v = glCreateShader(GL_VERTEX_SHADER);
  f = glCreateShader(GL_FRAGMENT_SHADER);
  
  vs = textFileRead("bumpearth.vert");
  fs = textFileRead("bumpearth.frag");
  
  const char * vv = vs;
  const char * ff = fs;
  
  glShaderSource(v, 1, &vv,NULL);
  glShaderSource(f, 1, &ff,NULL);
  
  free(vs);free(fs);
  
  glCompileShader(v);
  glCompileShader(f);
  
  printShaderInfoLog(v);
  printShaderInfoLog(f);
  
  p = glCreateProgram();
  glAttachShader(p,v);
  glAttachShader(p,f);
  
  glLinkProgram(p);
  printProgramInfoLog(p);
  
  glUseProgram(p);

  // set initial value

  glUniform1f(glGetUniformLocation(p, "Exaggeration"), exaggeration_factor);
}

//----------------------------------------------------------------------------

void load_textures(int numTextures, char **texfilename)
{
  int i;
  unsigned char **texim;
  int *texim_w, *texim_h;

  texim = (unsigned char **) calloc(numTextures, sizeof(unsigned char *));
  texim_w = (int *) calloc(numTextures, sizeof(int));
  texim_h = (int *) calloc(numTextures, sizeof(int));

  for (i = 0; i < numTextures; i++) {

    texim[i] = read_PPM(texfilename[i], &texim_w[i], &texim_h[i]);

    printf("loading %s: w = %i, h = %i\n", texfilename[i], texim_w[i], texim_h[i]);

    glBindTexture(GL_TEXTURE_2D, i+1);

    glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
    glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);

    glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
    glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);

    glTexImage2D(GL_TEXTURE_2D, 0, 4, texim_w[i], texim_h[i], 0, GL_RGBA, 
		 GL_UNSIGNED_BYTE, texim[i]);
  }
}

//----------------------------------------------------------------------------

// NO pixels are transparent

unsigned char *read_PPM(char *filename, int *width, int *height)
{
  int i, j, w, h;
  FILE *fp;
  int r, g, b, t;
  bool iscomment;
  unsigned char ur, ug, ub;
  char imtype[3];
  char str[80];
  char c;
  unsigned char *im;
  int bpp;

  // get size info

  fp = fopen(filename, "rb");
  if (!fp) {
    printf("read_PPM(): no such file");
    exit(1);
  }

  fscanf(fp, "%s\n", (char *) (&imtype));

  // attempt to eat comments

  do {
    iscomment = false;
    c = fgetc(fp);
    ungetc(c, fp);
    if (c == '#') {
      iscomment = true;
      fgets (str, 79, fp);
    }
  } while (iscomment);

  // read image dimensions

  fscanf(fp, "%i %i\n255\n", &w, &h);    
  *width = w;
  *height = h;

  // allocate image

  bpp = 4;

  im = (unsigned char *) calloc(w * h * bpp, sizeof(unsigned char));

  // actually read it in

  // ascii

  if (!strcmp(imtype, "P3")) {

    for (j = 0, t = 0; j < h; j++) {
      for (i = 0; i < w; i++, t += bpp) {
	fscanf(fp, "%i %i %i ", &r, &g, &b);
	im[t] = (unsigned char) r;
	im[t+1] = (unsigned char) g;
	im[t+2] = (unsigned char) b;
// 	if (r == 0 && g == 0 && b == 0)
// 	  im[t+3] = (unsigned char) 0;
// 	else
	  im[t+3] = (unsigned char) 255;
      }
    }
    fscanf(fp, "\n");
  }

  // binary

  else if (!strcmp(imtype, "P6")) {
    
    for (j = 0, t = 0; j < h; j++) {
      for (i = 0; i < w; i++, t += bpp) {
	fscanf(fp, "%c%c%c", &ur, &ug, &ub);
	im[t] = (unsigned char) ur;
	im[t+1] = (unsigned char) ug;
	im[t+2] = (unsigned char) ub;
// 	if (ur == 0 && ug == 0 && ub == 0)
// 	  im[t+3] = (unsigned char) 0;
// 	else
	  im[t+3] = (unsigned char) 255;
      }
    }
  }

  // unknown

  else {
    printf("unrecognized ppm file type");
    exit(1);
  }

  // finish up

  fclose(fp);

  return im;
}

//----------------------------------------------------------------------------

void initializeTextures()
{
  int numTextures = 5;
  char **texfilename = (char **) malloc(numTextures * sizeof(char *));
  for (int i = 0; i < numTextures; i++) 
    texfilename[i] = (char *) malloc(256 * sizeof(char));

  sprintf(texfilename[0], "big_earth.ppm");
  sprintf(texfilename[1], "big_water.ppm");
  sprintf(texfilename[2], "big_night.ppm");
  sprintf(texfilename[3], "big_clouds.ppm");
  sprintf(texfilename[4], "big_elevation.ppm");

  load_textures(numTextures, texfilename);

  qObj = gluNewQuadric();
  gluQuadricNormals(qObj, GLU_SMOOTH);
  gluQuadricTexture(qObj, GL_TRUE);

  glEnable(GL_TEXTURE_2D);

}

//----------------------------------------------------------------------------

int main(int argc, char **argv) 
{
  glutInit(&argc, argv);
  glutInitDisplayMode(GLUT_DEPTH | GLUT_DOUBLE | GLUT_RGBA);
  glutInitWindowPosition(100,100);
  glutInitWindowSize(320,320);
  glutCreateWindow("Earth");
  
  glutDisplayFunc(renderScene);
  glutIdleFunc(renderScene);
  glutReshapeFunc(changeSize);
  glutKeyboardFunc(keyboard);
  
  glEnable(GL_DEPTH_TEST);
  glClearColor(1.0,1.0,1.0,1.0);
  
  glewInit();
  if (glewIsSupported("GL_VERSION_2_0"))
    printf("Ready for OpenGL 2.0\n");
  else {
    printf("OpenGL 2.0 not supported\n");
    exit(1);
  }

  initializeTextures();

  setShaders();
  
  glutMainLoop();
  
  return 0;
}

//----------------------------------------------------------------------------
//----------------------------------------------------------------------------
